   
      program select_vNsds
c*************************************************************************
c
c    Last modified February 10, 1999
c
c    Program to read in a TRMM HDF-EOS data file, and write out 
c    user-selected SDS arrays and Vdata tables as separate flat binary 
c    files. Metadata (global attributes) are written out in a separate 
c    ascii file with extension .meta
c
c    Data in output files will be the same number type (float32, int32, 
c    int16, int8, uint8) as the corresponding array in the input (HDF) 
c    file. 
c
c
c    USAGE :
c
c    select_vNsds <trmm_hdf_file>
c    
c
c    NOTES :
c
c    1) On output, record length is specified as 4 byte words (for SGI);
c       For SUN machines, please modify OPEN statements to reflect record
c       length as number of bytes
c
c    2) Please make sure that the buffer arrays are dimensioned properly
c       via the PARAMETER statements
c
c    3) Output filenames automatically created by replacing 'HDF' suffix
c       of input file with selected parameter (SDS) names
c
c*************************************************************************

      parameter   ( maxscn   = 18200 ,  maxpix    = 261 )
      parameter   ( maxchn   =    20 ,  maxdim    =   6 )
      parameter   ( maxchar  =    20 ,  maxchrfn  =  80 )  

      parameter   ( maxsds   =   200 )  

      parameter   ( maxchrga = 20000 ,  maxrecsiz =  88 )
      parameter   ( maxsdsiz = maxscn * maxpix * maxchn )
      parameter   ( maxvdsiz = maxscn * maxrecsiz       )
      parameter   ( maxvdata = 200 )
   
      character * (maxchrfn)             infile
      character * (maxchrfn + maxchar)   outfile

      integer     file_id  , status
      integer     hopen    , hclose

      integer     data_type , attr_index, count , size
      integer     nglobal_attr , nsds_attr  
      character * (maxchar)   attr_name
      character * (maxchrga)  attr_val   

      integer    sds_id  , sd_id
      integer    sfgainfo, sffinfo , sfstart , sfginfo , sfselect
      integer    sfendacc, sfdimid , sfgdinfo, sfrdata , sfrcdata
      integer    sfgcal  , sfrattr , sfscatt , sfend

      character * (maxchar)  sds_name 

      integer    vfstart , vsfgid  , vsfatch , vsfinq  , vsfread 
      integer    vfend   , vsffdef , vsfsfld , vsfsnam , vsfscls
      integer    vsfdtch , vsfrd   , vsfrdc 
      integer    vfnflds , vffisiz , vffordr , vfftype , vffname 
      integer    vdata_ref , vdata_id , vdata_id1 , vdata_size
      integer    interlace , num_rec  , nrec_rd   , num_fields

      integer    vdata_index(maxvdata), total_vref(maxvdata)
      integer    total_vindex


      character * 9           substring
      character * 9           vdata_substr(3)
      data                    vdata_substr /   'navigatio' ,
     .                                         'scan_stat' ,
     .                                         'scan_time'  /
      logical                 vdata_write

      character * (maxvdsiz)  vdata 
      character * (maxchar )  vdata_name 

      character * (1024 )  field_name, fields
      
      integer    field_order, field_type, field_size, field_index

      integer  DFACC_READ, DFACC_WRITE, DFACC_CREATE
      data     DFACC_READ     / 1 /
      data     DFACC_WRITE    / 2 /
      data     DFACC_CREATE   / 4 /

      integer  FULL_INTERLACE
      data     FULL_INTERLACE / 0 /

      integer*1  i1data  ( maxchn * maxpix * maxscn )
      integer*2  i2data  ( maxchn * maxpix * maxscn )
      integer*4  i4data  ( maxchn * maxpix * maxscn )      
      real*4     r4data  ( maxchn * maxpix * maxscn )
      character  cdata   ( maxchn * maxpix * maxscn ) 

      equivalence ( i2data, r4data ), ( cdata, i2data )
      equivalence ( i4data, r4data ), ( cdata, i1data )

      integer    dimsizes(maxdim),  rank
      integer    start(maxdim)   ,  stride(maxdim)  , edges(maxdim)      

      real*8     gain , gain_err , offset , offset_err
      integer*4  cal_data_type
      integer*2  exp_bitg , exp_bito 
      equivalence ( gain, exp_bitg ), (offset, exp_bito) 

      integer*2  Z7FFF
      data       Z7FFF / 32767 /

      integer*4  indx_sds(maxsds)
      data       indx_sds  / maxsds * -99 /

c
c  Check for input file name specification
c
      iargs = iargc()
      if ( iargs .lt. 1 )  then
          print * , '  '
          print * , ' Usage: read_trmm trmm_hdf_file '
          print * , '  '
          stop
      endif

      call getarg ( 1, infile )

c
c  Locate position of 'HDF' suffix in input file name 
c
      iprefix = -1
      do i = 1, maxchrfn - 2
        if ( (infile(i:i+2) .eq. 'HDF') .or.
     .       (infile(i:i+2) .eq. 'hdf') )   iprefix = i - 1
      enddo

      if ( iprefix .lt. 0 )  then
         print * , '  '
         print * , ' Input file does not have an HDF extension'
         print * , '  '
         stop
      endif

c
c  Open connections to original (input) TRMM HDF file 
c
      file_id  = hopen(infile, DFACC_READ, 0)
      if ( file_id .eq. -1 )  then
         print * , '  '
         print * , ' Error opening input file : ', infile
         stop
      endif    

c
c  Open HDF SDS and Vdata Interfaces
c
      sd_id    = sfstart(infile, DFACC_READ)
      if ( sd_id .eq. -1 ) then
         print * , '  '
         print * , ' Error opening HDF SD interface '
         stop
      endif

      status   = vfstart(file_id)
      if ( status .eq. -1 ) then
         print * , '  '
         print * , ' Error opening HDF Vdata interface '
         stop
      endif

c
c  Obtain number of SDSs and Global Attributes in input file
c
      status = sffinfo(sd_id, nsds, nglobal_attr)
      if ( status .ne. 0 ) then
         print * , '  '
         print * , ' Error return from HDF routine sffinfo  '
         stop
      endif 

c
c  Read in global attributes which should contain Core and Archive 
c  metadata - write out to ascii file
c
      outfile = ' '
      outfile(1:iprefix) = infile(1:iprefix)
      outfile(iprefix+1:iprefix+4) = 'meta'

      open(unit = 10, file = outfile, form='formatted')

      print * , '  '
      print * , ' Contents of TRMM HDF file : ', infile
      print * , '  '                                
      print * , '  ' 
      print * , ' Number of global attr in file = ' , nglobal_attr 
      print * , '  '

      do i = 1, nglobal_attr

         attr_index = i - 1
         status = sfgainfo(sd_id, attr_index, attr_name, data_type,
     .                       count)
         if ( status .ne. 0 )  then
             print * , ' '
             print * , ' Error return from HDF routine sfgainfo '
             stop
         endif 
         print * , ' attr name  = ' , attr_name
         print * , ' attr type  = ' , data_type
         print * , ' attr count = ' , count
         status  =  sfrattr(sd_id, attr_index, attr_val)
         if ( status .ne. 0 )  then
             print * , ' '
             print * , ' Error return from HDF routine sfrattr '
             stop
         endif  

         write(10,*)  attr_val(1:count)
         write(10,*)  ' '
         write(10,*)  ' '
      enddo

      close (10)

c
c  List out SDSs contained in input HDF file
c
      print * , '  ' 
      print * , ' Number of SDS arrays in file = ' , nsds
      print * , '  '

      do n = 1, nsds
         sds_name = ' '
         sds_id   = sfselect(sd_id, n-1)
         status = sfginfo(sds_id, sds_name, rank, dimsizes, data_type,
     .                                 nsds_attr)
         if ( status .ne. 0 ) then
            print * , '  '
            print * , ' Error return from HDF routine sfginfo '
            stop
         endif         
c
c  Check that rank of SDS is less than MAXDIM
c
         if ( rank .gt. maxdim )  then 
             print * , ' '
             print * , ' RANK OUT OF RANGE; INCREASE MAXDIM TO:', rank
             print * , ' '
             stop
         endif

         write(6,100)  n, sds_name, ( dimsizes(i), i= 1, rank )
  100    format( 1x, i3, ')', 1x, a, 3x, 'dimensions = ', 6i6 )

      enddo

c
c   Prompt user for desired parameters to be written as binary file 
c
   99 print  * , ' '
      print  * , ' Enter total number of parameters to write out '
      print  * , '            or 0 to exit program               '
      read(5,*),   nparm

      if ( nparm .le. 0 )   then
           print * , ' No program selected - # of parameters .le. 0 '
      elseif ( nparm .eq. nsds )  then
           do i = 1, nsds
              indx_sds(i) = i
           enddo    
      else                
           print  *  , ' Enter parameter numbers from above list,'
           print  *  , ' separated by white space '
           read(5,*) ,   (indx_sds(n), n = 1, nparm)
      endif

c
c   Check for valid parameter input
c
      if ( nparm .gt. nsds )  then
         print * , '  ' 
         print * , ' Invalid entry for total number of parameters '
         go to 99
      endif

      do n = 1, nparm
         if ( (indx_sds(n) .gt. nsds) .or. (indx_sds(n) .lt. 1) )  then
              print * , '  '
              print * , ' Invalid parameter index specified '
              go to 99
         endif
      enddo
       
c
c   Now loop through SDSs and write selected array(s) to binary files
c

      do n = 1, nparm

             nn = indx_sds(n) 
             sds_name = ' '
             sds_id   = sfselect(sd_id, nn-1)
             status = sfginfo(sds_id, sds_name, rank, dimsizes, 
     .                        data_type, nsds_attr)  
         
             status = sfgcal(sds_id, gain, gain_err, offset, 
     .                       offset_err, cal_data_type )
c
c  Check for nan in case HDF calibration tags were not initialized

             if ( exp_bitg .eq. Z7FFF )  gain   = 0.0
             if ( exp_bito .eq. Z7FFF )  offset = 0.0

             print * , '****************************************'  
             print * , ' SDS name  = ',  sds_name
             print * , ' SDS type  = ',  data_type
             print * , ' SDS rank  = ',  rank
             print * , ' SDS scale = ',  gain
             print * , ' SDS offset= ',  offset
             print * , ' SDS dims  = ', (dimsizes(j), j =1, rank)

             do j = 1, rank
                edges(j)  = dimsizes(j)
                stride(j) = 1
                start(j)  = 0
             enddo

             num_elements = 1
             do j = 1, rank
                num_elements = num_elements * dimsizes(j)
             enddo
c
c  Check that total SDS size is less than MAXSIZ
c
             if ( num_elements .gt. maxsdsiz )  then
                print * , ' '
                print * , 'DIMENSIONS OUT OF RANGE; INCREASE BUFFERSIZE'
                print * , ' '
                stop
             endif


             outfile = ' '
             outfile(1:iprefix) = infile(1:iprefix)
             outfile(iprefix+1:iprefix+maxchar)= sds_name(1:maxchar) 
c
c   Read TRMM floating point arrays ....
c
             if ( data_type .eq. 5 ) then 
                 status  = sfrdata(sds_id, start, stride, edges, r4data)
                 if ( status .ne. 0 ) then
                    print * , '  '
                    print * , ' Error return from HDF routine sfrdata '
                    stop
                 endif         
             
                 lenrec = num_elements
                 open( 10, file=outfile, status='unknown', recl=lenrec,
     .                 access='direct', form='unformatted' )
                 write(10) (r4data(k), k = 1, num_elements)
                 close(10)

c
c   Read TRMM 32 bit integer arrays....
c
             elseif ( data_type .eq. 24 ) then
                 status = sfrdata(sds_id, start, stride, edges, i4data)
                 if ( status .ne. 0 ) then
                    print * , '  '
                    print * , ' Error return from HDF routine sfrdata '
                    stop
                 endif

                 lenrec = num_elements
                 open( 10, file=outfile, status='unknown', recl=lenrec,
     .                 access='direct', form='unformatted' )
                 write(10) (i4data(k), k= 1, num_elements)
                 close(10)

c
c   Read TRMM 16 bit integer arrays....
c
             elseif ( data_type .eq. 22 ) then 
                 status = sfrdata(sds_id, start, stride, edges, i2data)
                 if ( status .ne. 0 ) then
                    print * , '  '
                    print * , ' Error return from HDF routine sfrdata '
                    stop
                 endif         

                 irem    = mod(num_elements * 2, 4)
                 lenrec  = num_elements * 2 / 4
                 if (irem .ne. 0) lenrec = lenrec + 1         

                 open( 10, file=outfile, status='unknown', recl=lenrec,
     .                 access='direct', form='unformatted' )
                 write(10) (i2data(k), k= 1, num_elements)
                 close(10) 

c
c   Read TRMM 8-bit signed and unsigned integer arrays....
c
             elseif( (data_type .eq. 20) .or. (data_type .eq. 21) ) then
                 status = sfrdata(sds_id, start, stride, edges, cdata)
                 if ( status .ne. 0 ) then
                    print * , '  '
                    print * , ' Error return from HDF routine sfrdata '
                    stop
                 endif

                 irem    = mod(num_elements , 4)
                 lenrec  = num_elements / 4
                 if (irem .ne. 0) lenrec = lenrec + 1

                 open( 10, file=outfile, status='unknown', recl=lenrec,
     .                 access='direct', form='unformatted' )
                 write(10) (cdata(k), k= 1, num_elements)
                 close(10)

              else
                 print * , ' Unexpected number type encountered '
                 stop
 
              endif       

              print * , ' '
              print * , ' Binary file successfully created : ', outfile

c
c   Close access to current SDS
c 
              status = sfendacc(sds_id )

      enddo

c
c   Close SDS interface
c    
      status = sfend(sd_id )  

c
c   Now read in the Vdata tables and write non-dimension Vdatas to 
c   separate output files
c      
      vdata_ref = - 1
      vdata_ref =  vsfgid(file_id, vdata_ref)
      if ( vdata_ref .eq. -1 )  then
           print * , ' No Vdatas found in data set '
           stop
      endif

cb=======================================================

      print *, ' '
      print *, ' '
      print *, ' '
      print *, ' List of Vdata name:              (rec_num   rec_size)'
      print *, ' '

      total_vindex = 0
      do while ( (vdata_ref .ne. -1) )

          vdata_id    = vsfatch(file_id, vdata_ref, 'r')
          vdata_name  = ' '
          fields      = ' '
          status      = vsfinq(vdata_id, num_rec, interlace, fields,
     .                         vdata_size, vdata_name )
          if ( status .ne. 0 )  then
              print * , ' '
              print * , ' Error return from HDF routine vsfinq '
              stop
          endif

          if ( vdata_name(1:7) .ne. "fakeDim" .and.
     .         vdata_name(1:10) .ne. "add_offset" .and. 
     .         vdata_name(1:12) .ne. "scale_factor") then 

               total_vindex = total_vindex + 1
               total_vref(total_vindex) = vdata_ref

               write(6, 1100) total_vindex, vdata_name, 
     .                        num_rec, vdata_size
 1100          format(1x, i3, 3x, a20, 12x, i5,  5x, i5)
          endif

          status    = vsfdtch( vdata_id )
          vdata_ref = vsfgid ( file_id, vdata_ref )

      enddo

c 
c  Prompt user for desired vdata to be written as binary file
c 

      print  *  , ' '
      print  *  , ' '
      print  * , ' Enter total number of vdata to write out '
      print  * , '            or 0 to exit program               '
      read(5,*),   nparm

      if ( nparm .le. 0 )   then
           vdata_index(1) = 0
      elseif ( nparm .eq. total_vindex )  then
           do i = 1, total_vindex
              vdata_index(i) = i 
           enddo    
      else                
           print  *  , ' Enter parameter numbers from above list,'
           print  *  , ' separated by white space '
           read(5,*) ,   (vdata_index(n), n = 1, nparm)
      endif

      if (nparm .le. 0) then
          print * , ' No vdata selected - # of vdata .le. 0 '
      else 
          print * , ' The valid numbers you enter are: '
          print * , (vdata_index(n) , n = 1, nparm)

      endif

c 
c  Now loop through vdatas and write selected data(s) to binary files
c 

      if(nparm .gt. 0) then
         do n = 1, nparm
            nindex = vdata_index(n)
            vdata_id    = vsfatch(file_id, total_vref(nindex), 'r')
            vdata_name  = ' '
            fields      = ' '
            status      = vsfinq(vdata_id, num_rec, interlace, fields,
     .                           vdata_size, vdata_name )
            if ( status .ne. 0 )  then
                print * , ' '
                print * , ' Error return from HDF routine vsfinq '
                stop
            endif

            print * , ' *************************************'
            print * , ' Vdata name     :     ' , vdata_name(1:20)
            print * , ' Vdata recs     :     ' , num_rec
            print * , ' Vdata recsize  :     ' , vdata_size
 
            status      = vsfsfld( vdata_id, fields )
            nrec_rd     = vsfrdc ( vdata_id, vdata, num_rec, 
     .                                       FULL_INTERLACE ) 

            if (nrec_rd .ne. num_rec )  then 
                print * , ' '
                print * , ' Vdata recs read and num_rec do not match'
                stop 
            endif

            num_fields = vfnflds(vdata_id)

            print * , ' Field Name list:     ' 
            do i = 1, num_fields
               field_index  = i - 1
               field_name   = ' '
               status = vffname(vdata_id, field_index, field_name)
               field_size   = vffisiz( vdata_id, field_index )
               field_order  = vffordr( vdata_id, field_index )
               field_type   = vfftype( vdata_id, field_index )
               write(6, 1200) field_name, field_size , field_type
 1200          format(1x, 6x, a20, '(size: ', i5, '   type:', i3, ')')
            enddo

c
c   Now write out the Vdata information to a binary file
c
            outfile = ' '
            outfile(1:iprefix) = infile(1:iprefix)
            outfile(iprefix+1:iprefix+maxchar)= vdata_name(1:maxchar)   

            numchars = num_rec * vdata_size
            irem     = mod(numchars , 4) 
            lenrec   = numchars / 4
            if (irem .ne. 0) lenrec = lenrec + 1
 
            open( 10, file=outfile, status='unknown', recl=lenrec,
     .            access='direct', form='unformatted' )
            write (10)  vdata(1:numchars)
            close (10)

            print * , ' ' 
            print * , ' Binary file successfully created : ', outfile

            status    = vsfdtch( vdata_id )

         enddo
      endif

      print * , ' '
      print * , ' '

      status = vfend ( file_id )
c
c   Close input file
c
      status = hclose( file_id )

      stop 
      end

